from django.contrib import admin

from ..models.agreements import AgreementOption


class AgreementOptionInlineAdmin(admin.TabularInline):
    model   = AgreementOption
    extra   = 0
    min_num = 1


class AgreementAdmin(admin.ModelAdmin):
    inlines         = (AgreementOptionInlineAdmin,)
    list_display    = ('name', 'order')
    list_editable   = ('order',)



import locale
import os
import string
import unicodedata
import zlib
from datetime import date
from urllib.parse import urlencode

from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse_lazy as reverse
from django.db import IntegrityError, transaction
from django.utils.encoding import iri_to_uri, smart_text
from django.utils.translation import ugettext_lazy as _

from .conf import settings


def _get_localeconv():
    """
    This function loads localeconv during module load.
    It is necessary, because using locale.setlocale later may be dangerous
    (It is not thread-safe in most of the implementations.)
    """
    original_locale_name = locale.setlocale(locale.LC_ALL)
    locale_name = locale.locale_alias[settings.LANGUAGE_CODE].split('.')[0] + '.UTF-8'
    locale.setlocale(locale.LC_ALL, str(locale_name))
    lc = locale.localeconv()
    locale.setlocale(locale.LC_ALL, original_locale_name)
    return lc


localeconv = _get_localeconv()


# This function is inspired by python's standard locale.currency().
def currency(val, international=False):
    """Formats val according to the currency settings for current language."""
    digits = settings.PRICE_DECIMAL_PLACES

    # grouping
    groups = []
    s = str(abs(int(val)))
    for interval in locale._grouping_intervals(localeconv['mon_grouping']):
        if not s:
            break
        groups.append(s[-interval:])
        s = s[:-interval]
    if s:
        groups.append(s)
    groups.reverse()
    s = smart_text(localeconv['mon_thousands_sep']).join(groups)

    # display fraction for non integer values
    if digits and not isinstance(val, int):
        s += smart_text(localeconv['mon_decimal_point']) + '{{:.{}f}}'.format(digits).format(val).split('.')[1]

    # '<' and '>' are markers if the sign must be inserted between symbol and value
    s = '<' + s + '>'

    smb = smart_text(localeconv[international and 'int_curr_symbol' or 'currency_symbol'])
    precedes = localeconv[val < 0 and 'n_cs_precedes' or 'p_cs_precedes']
    separated = localeconv[val < 0 and 'n_sep_by_space' or 'p_sep_by_space']

    if precedes:
        s = smb + (separated and ' ' or '') + s
    else:
        s = s + (separated and ' ' or '') + smb

    sign_pos = localeconv[val < 0 and 'n_sign_posn' or 'p_sign_posn']
    sign = localeconv[val < 0 and 'negative_sign' or 'positive_sign']

    if sign_pos == 0:
        s = '(' + s + ')'
    elif sign_pos == 1:
        s = sign + s
    elif sign_pos == 2:
        s = s + sign
    elif sign_pos == 3:
        s = s.replace('<', sign)
    elif sign_pos == 4:
        s = s.replace('>', sign)
    else:
        # the default if nothing specified;
        # this should be the most fitting sign position
        s = sign + s

    return s.replace('<', '').replace('>', '').replace(' ', '\u00A0')


def amount_color(amount):
    if amount > 0:
        return settings.LEPRIKON_COLOR_POSITIVE
    elif amount < 0:
        return settings.LEPRIKON_COLOR_NEGATIVE
    else:
        return settings.LEPRIKON_COLOR_ZERO


def comma_separated(lst):
    lst = list(map(smart_text, lst))
    if len(lst) > 2:
        return _(', and ').join([', '.join(lst[:-1]), lst[-1]])
    else:
        return _(', and ').join(lst)


def get_rand_hash(length=32, stringset=string.ascii_letters + string.digits):
    return ''.join([stringset[i % len(stringset)] for i in [ord(x) for x in os.urandom(length)]])


def current_url(request):
    if request.META['QUERY_STRING']:
        return '{}?{}'.format(request.path, request.META['QUERY_STRING'])
    else:
        return request.path


def url_back(request):
    return request.POST.get(
        settings.LEPRIKON_PARAM_BACK,
        request.GET.get(
            settings.LEPRIKON_PARAM_BACK,
            reverse('leprikon:summary'),
        ),
    )


def url_with_back(url, url_back):
    return '{}?{}'.format(url, urlencode({settings.LEPRIKON_PARAM_BACK: iri_to_uri(url_back)}))


def reverse_with_back(request, *args, **kwargs):
    return url_with_back(reverse(*args, **kwargs), current_url(request))


def get_birth_date(birth_num):
    birth_num = birth_num.replace('/', '')
    y = int(birth_num[:2])
    if len(birth_num) == 9:
        # before 1954
        if y < 54:
            year = 1900 + y
        else:
            year = 1800 + y
    else:
        year = int(date.today().year / 100) * 100 + y
        if y > date.today().year % 100:
            year -= 100
    month = int(birth_num[2:4]) % 50 % 20
    day = int(birth_num[4:6])
    return date(year, month, day)


def get_age(birth_date, today=None):
    today = today or date.today()
    if date(today.year, birth_date.month, birth_date.day) > today:
        return today.year - birth_date.year - 1
    else:
        return today.year - birth_date.year


def first_upper(s):
    return s[0].upper() + s[1:] if s else ''


def merge_objects(source, target, attributes=None, exclude=[]):
    attributes = attributes or [f.name for f in source._meta.fields if f.name not in exclude]
    for attr in attributes:
        if not getattr(target, attr):
            setattr(target, attr, getattr(source, attr))
    return target


@transaction.atomic
def merge_users(source, target):
    target = merge_objects(source, target, ('first_name', 'last_name', 'email'))
    target.date_joined = min(source.date_joined, target.date_joined)
    target.last_login = max(source.last_login, target.last_login)

    try:
        leader = source.leprikon_leader
        leader.user = target
        leader.save()
    except ObjectDoesNotExist:
        pass
    except IntegrityError:
        # both users are leaders
        raise

    source.leprikon_registrations.update(user=target)

    for sp in source.leprikon_participants.all():
        tp = target.leprikon_participants.filter(birth_num=sp.birth_num).first()
        if tp:
            tp = merge_objects(sp, tp, exclude=('id', 'user', 'birth_num'))
            tp.save()
        else:
            sp.user = target
            sp.save()

    for sp in source.leprikon_parents.all():
        tp = target.leprikon_parents.filter(first_name=sp.first_name, last_name=sp.last_name).first()
        if tp:
            tp = merge_objects(sp, tp, exclude=('id', 'user'))
            tp.save()
        else:
            sp.user = target
            sp.save()

    for mr in source.leprikon_messages.all():
        if not target.leprikon_messages.filter(message=mr.message).exists():
            mr.recipient = target
            mr.save()

    try:
        # support social auth
        source.social_auth.update(user=target)
    except AttributeError:
        pass

    source.delete()


def spayd(*items):
    s = 'SPD*1.0*' + '*'.join(
        '%s:%s' % (k, unicodedata.normalize('NFKD', str(v).replace('*', '')).encode('ascii', 'ignore').upper().decode())
        for k, v in sorted(items)
    )
    s += '*CRC32:%x' % zlib.crc32(s.encode())
    return s.upper()



from . import (  # noqa
    account, agegroup, agreements, citizenship, courses, department, events,
    fields, leprikonsite, messages, place, printsetup, question, roles, school,
    schoolyear, startend, subjects, timesheets, useragreement, utils,
)



from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext_lazy as _

from ..forms.courses import (
    CourseJournalEntryForm, CourseJournalLeaderEntryForm,
)
from ..models.courses import (
    Course, CourseJournalEntry, CourseJournalLeaderEntry,
)
from .generic import (
    CreateView, DeleteView, DetailView, TemplateView, UpdateView,
)


class CourseAlternatingView(TemplateView):
    template_name = 'leprikon/course_alternating.html'

    def get_title(self):
        return _('Alternating in school year {}').format(self.request.school_year)

    def get_context_data(self, **kwargs):
        context = super(CourseAlternatingView, self).get_context_data(**kwargs)
        context['alternate_leader_entries'] = self.request.leader.get_alternate_leader_entries(self.request.school_year)
        return context



class CourseRegistrationsView(DetailView):
    model = Course
    template_name_suffix = '_registrations'

    def get_queryset(self):
        qs = super(CourseRegistrationsView, self).get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leaders=self.request.leader)
        return qs



class CourseJournalView(DetailView):
    model = Course
    template_name_suffix = '_journal'

    def get_queryset(self):
        qs = super(CourseJournalView, self).get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leaders=self.request.leader)
        return qs



class CourseJournalEntryCreateView(CreateView):
    model           = CourseJournalEntry
    form_class      = CourseJournalEntryForm
    template_name   = 'leprikon/coursejournalentry_form.html'
    title           = _('New journal entry')
    message         = _('The journal entry has been created.')

    def dispatch(self, request, *args, **kwargs):
        if self.request.user.is_staff:
            self.course = get_object_or_404(Course, id=int(kwargs.pop('course')))
        else:
            self.course = get_object_or_404(Course, id=int(kwargs.pop('course')), leaders=self.request.leader)
        return super(CourseJournalEntryCreateView, self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super(CourseJournalEntryCreateView, self).get_form_kwargs()
        kwargs['course'] = self.course
        return kwargs



class CourseJournalEntryUpdateView(UpdateView):
    model           = CourseJournalEntry
    form_class      = CourseJournalEntryForm
    template_name   = 'leprikon/coursejournalentry_form.html'
    title           = _('Change journal entry')
    message         = _('The journal entry has been updated.')

    def get_object(self):
        obj = super(CourseJournalEntryUpdateView, self).get_object()
        if (self.request.user.is_staff or self.request.leader in obj.course.all_leaders + obj.all_alternates):
            return obj
        else:
            raise Http404()



class CourseJournalEntryDeleteView(DeleteView):
    model   = CourseJournalEntry
    title   = _('Delete journal entry')
    message = _('The journal entry has been deleted.')

    def get_queryset(self):
        return super(CourseJournalEntryDeleteView, self).get_queryset().filter(
            course__leaders = self.request.leader,
        )

    def get_object(self):
        obj = super(CourseJournalEntryDeleteView, self).get_object()
        if obj.timesheets.filter(submitted = True).exists():
            raise Http404()
        return obj

    def get_question(self):
        return _('Do You really want to delete course journal entry?')



class CourseJournalLeaderEntryUpdateView(UpdateView):
    model           = CourseJournalLeaderEntry
    form_class      = CourseJournalLeaderEntryForm
    template_name   = 'leprikon/coursejournalleaderentry_form.html'
    title           = _('Change timesheet entry')
    message         = _('The timesheet entry has been updated.')

    def get_object(self):
        obj = super(CourseJournalLeaderEntryUpdateView, self).get_object()
        if (
            self.request.user.is_staff or obj.timesheet.leader == self.request.leader or
            self.request.leader in obj.course_entry.course.all_leaders
        ):
            return obj
        else:
            raise Http404()



class CourseJournalLeaderEntryDeleteView(DeleteView):
    model   = CourseJournalLeaderEntry
    title   = _('Delete timesheet entry')
    message = _('The timesheet entry has been deleted.')

    def get_queryset(self):
        return super(CourseJournalLeaderEntryDeleteView, self).get_queryset().filter(
            timesheet__leader = self.request.leader,
            timesheet__submitted = False,
        )

    def get_question(self):
        return _('Do You really want to delete timesheet entry?')



from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from filer.fields.file import FilerFileField
from PyPDF2 import PdfFileReader
from reportlab.lib.pagesizes import A4, portrait
from reportlab.lib.units import mm


@python_2_unicode_compatible
class PrintSetup(models.Model):
    name        = models.CharField(_('name'), max_length=150)
    background  = FilerFileField(verbose_name=_('pdf background'), blank=True, null=True,
                                 on_delete=models.SET_NULL, related_name='+')
    top         = models.IntegerField(_('margin top'), blank=True, default=20,
                                      help_text=_('distance in milimetres'))
    left        = models.IntegerField(_('margin left'), blank=True, default=20,
                                      help_text=_('distance in milimetres'))
    right       = models.IntegerField(_('margin right'), blank=True, default=20,
                                      help_text=_('distance in milimetres'))
    bottom      = models.IntegerField(_('margin bottom'), blank=True, default=20,
                                      help_text=_('distance in milimetres'))

    class Meta:
        app_label           = 'leprikon'
        ordering            = ('name',)
        verbose_name        = _('print setup')
        verbose_name_plural = _('print setups')

    def __str__(self):
        return self.name

    @cached_property
    def background_pdf(self):
        return PdfFileReader(self.background.file) if self.background else None

    @cached_property
    def page_size(self):
        if self.background:
            mediaBox = self.background_pdf.getPage(0).mediaBox
            return [int(mediaBox[2]), int(mediaBox[3])]
        else:
            return portrait(A4)

    @cached_property
    def x1(self):
        return int(self.left * mm)

    @cached_property
    def y1(self):
        return int(self.bottom * mm)

    @cached_property
    def width(self):
        return int(self.page_size[0] - self.left * mm - self.right * mm)

    @cached_property
    def height(self):
        return int(self.page_size[1] - self.top * mm - self.bottom * mm)

    @cached_property
    def bill_y1(self):
        return int(self.page_size[1] / 2.0 + self.bottom * mm)

    @cached_property
    def bill_height(self):
        return int(self.page_size[1] / 2.0 - self.top * mm - self.bottom * mm)
